import re
from typing import Any, ClassVar, Optional, Type

import boto3
from pydantic import BaseModel, ConfigDict, Field


class TransformMixin:
    def transform(
        self,
        attribute: str,
        transformer: Optional[str],
    ) -> Any:
        """
        Transform an attribute using a regular expression into something else
        before it is returned.

        .. important::
            This only makes sense for attributes that are strings.

        ``transformer`` is a regular expression that will be used to transform
        the value of the attribute.

        * If the attribute is ``None``, it will be returned verbatim.
        * If ``transformer`` is ``None``, the attribute will be returned verbatim.
        * If ``transformer`` has no named groups, the attribute will be replaced
          with the value of the first group.
        * If ``transformer`` has named groups, the attribute will be replaced
          with a dictionary of the named groups.

        Raises:
            ValueError: If the attribute does not exist on the model.
            RuntimeError: If the transformer fails to match the attribute value.

        Args:
            attribute: The attribute to transform.
            transformer: The regular expression to use to transform the attribute.

        Returns:
            The transformed attribute.

        """
        if not hasattr(self, attribute):
            msg = f"Invalid attribute: {self.__class__.__name__}.{attribute}"
            raise ValueError(msg)
        if transformer is None:
            return getattr(self, attribute)
        value = getattr(self, attribute)
        if value is None:
            return None
        if match := re.search(transformer, value):
            if match.groupdict():
                return match.groupdict()
            return match.group(1)
        msg = (
            f"Transformer failed to match: transformer=r'{transformer}', "
            f"value='{getattr(self, attribute)}'"
        )
        raise RuntimeError(msg)


class Boto3Model(TransformMixin, BaseModel):
    """
    The base class for all boto3 models.
    """

    model_config = ConfigDict(validate_assignment=True)


class ReadonlyBoto3Model(Boto3Model):
    """
    The base class for all boto3 models that are readonly.
    """

    model_config = ConfigDict(frozen=True, validate_assignment=True, extra="allow")


class Boto3ModelManager(TransformMixin):
    #: The name of the boto3 service.  Example: ``ec2``, ``s3``, etc.
    service_name: str

    def __init__(self) -> None:
        #: The boto3 client for the AWS service
        self.client = boto3.client(self.service_name)  # type: ignore[call-overload]
        #: The boto3 session to use for this manager.
        self.session = boto3.session.Session()

    def using(self, session: boto3.session.Session) -> "Boto3ModelManager":
        """
        Use a different boto3 session for this manager.

        Args:
            session: The boto3 session to use.

        """
        # TODO: this is a bad way to do this -- it means that whatever session
        # was last set with .using() will be transparently used by all other
        # calls to the manager.  This is not good.  We need som way to make
        # it like a context manager, so that the session is only used for the
        # duration of the actual method call.
        self.session = session
        self.client = session.client(self.service_name)  # type: ignore[call-overload]
        return self

    def serialize(self, arg: Any) -> Any:
        """
        Some of our botocraft methods use :py:class:`Boto3Model` objects as
        arguments (e.g. ``create``, ``update``), but boto3 methods expect simple
        Python types.  This method will serialize the model into a set of types
        that boto3 will understand if it is a :py:class:`Boto3Model` object.

        While serializing, we always exclude ``None`` values, because boto3
        doesn't like them.

        If ``arg`` is not a :py:class:`Boto3Model` object, it will be returned
        verbatim.

        Args:
            arg: the botocraft method argument to serialize.

        Returns:
            A properly serialized argument.

        """
        if arg is None:
            return None
        if isinstance(arg, Boto3Model):
            return arg.model_dump(exclude_none=True)
        if isinstance(arg, list):
            # Oop, this is a list.  We need to serialize each item in the list.
            return [self.serialize(a) for a in arg]
        return arg

    def sessionize(self, response: Any) -> None:  # noqa: PLR0912
        """
        Look through ``response`` for any object with ``set_session`` as
        an attribute and set the session on that object.

        .. note::

            I'm making an assumption here that the only PrimaryBoto3Model
            or ReadonlyPrimaryBoto3Model objects that will be in the response
            will be at the top level.  If they are nested, we'll miss the ones
            that are nested.

        Args:
            response: The response to search.
        """
        if response is not None:
            if isinstance(response, BaseModel):
                # Pydantic models are ALWAYS iterable, so we can't use
                # iter(response) to check if it's a list.
                # We get the fields on the model, excluding our special
                # fields and then try to sessionize them.
                attrs = [
                    attr
                    for attr in response.__fields__
                    if attr not in ["session", "objects"]
                ]
                for attr in attrs:
                    _attr = getattr(response, attr)
                    if _attr is not None:
                        if isinstance(_attr, BaseModel):
                            if hasattr(_attr, "set_session"):
                                _attr.set_session(self.session)
                        else:
                            try:
                                iter(_attr)
                            except TypeError:
                                if hasattr(_attr, "set_session"):
                                    _attr.set_session(self.session)
                            else:
                                if _attr:
                                    if hasattr(_attr[0], "set_session"):
                                        for obj in _attr:
                                            obj.set_session(self.session)  # type: ignore[attr-defined]
            else:
                # This is NOT a pydantic model, now we test for iterability
                try:
                    iter(response)
                except TypeError:
                    # This is not an iterable
                    if hasattr(response, "set_session"):
                        response.set_session(self.session)  # type: ignore[attr-defined]
                else:
                    # This is an iterable
                    if hasattr(response[0], "set_session"):
                        [obj.set_session(self.session) or obj for obj in response]

    def get(self, *args, **kwargs):
        raise NotImplementedError

    def list(self, *args, **kwargs):
        raise NotImplementedError

    def create(self, model, **kwargs):
        raise NotImplementedError

    def update(self, model, **kwargs):  # noqa: ARG002
        # Some models cannot be updated, so instead of raising a
        # NotImplementedError, we raise a RuntimeError.
        msg = "This model cannot be updated."
        raise RuntimeError(msg)

    def delete(self, pk: str, **kwargs):
        raise NotImplementedError

    def get_waiter(self, name: str) -> Any:
        """
        Get a boto3 waiter object for this service.

        Args:
            name: The name of the waiter to get.

        Returns:
            The boto3 waiter object.

        """
        return self.client.get_waiter(name)


class ReadonlyBoto3ModelManager(Boto3ModelManager):  # pylint: disable=abstract-method
    def create(self, model, **kwargs):  # noqa: ARG002
        msg = "This model cannot be created."
        raise RuntimeError(msg)

    def update(self, model, **kwargs):  # noqa: ARG002
        msg = "This model cannot be updated."
        raise RuntimeError(msg)

    def delete(self, pk: str, **kwargs):  # noqa: ARG002
        msg = "This model cannot be deleted."
        raise RuntimeError(msg)


class ModelIdentityMixin:
    @property
    def pk(self) -> Optional[str]:
        """
        Get the primary key of the model instance.

        Returns:
            The primary key of the model instance.

        """
        raise NotImplementedError

    @property
    def arn(self) -> Optional[str]:
        """
        Get the ARN of the model instance.

        Returns:
            The ARN of the model instance.

        """
        msg = "The model does not have an ARN."
        raise ValueError(msg)

    @property
    def name(self) -> Optional[str]:
        """
        Get the name of the model instance.

        Returns:
            The name of the model instance.

        """
        msg = "The model does not have a name."
        raise ValueError(msg)


def classproperty(func):
    return classmethod(property(func))


class ReadonlyPrimaryBoto3Model(  # pylint: disable=abstract-method
    ModelIdentityMixin, ReadonlyBoto3Model
):
    #: The manager for this model
    manager_class: ClassVar[Type[Boto3ModelManager]]

    #: The boto3 session to use for this model.  This is set by the manager,
    #: and is used in relationships.  We have to use ``Any`` here because we
    #: pydantic complains vociferously if we use ``boto3.session.Session``.
    #: We exclude it from the model dump because it's not something that should
    #: be serialized.
    session: Optional[Any] = Field(default=None, exclude=True)

    @classproperty
    def objects(cls) -> Boto3ModelManager:  # noqa: N805
        """
        Get the manager for this model.

        Returns:
            The manager for this model.

        """
        return cls.manager_class()

    def save(self, **kwargs):  # noqa: ARG002
        """
        Save the model.
        """
        msg = "Cannot save a readonly model."
        raise RuntimeError(msg)

    def delete(self):
        """
        Delete the model.
        """
        msg = "Cannot delete a readonly model."
        raise RuntimeError(msg)

    def set_session(self, session: boto3.session.Session) -> None:
        """
        Set the boto3 session for this model.

        Args:
            session: The boto3 session to use.

        Returns:
            The model instance.

        """
        self.model_config["frozen"] = False
        self.session = session  # type: ignore[misc]
        self.model_config["frozen"] = True


class PrimaryBoto3Model(  # pylint: disable=abstract-method
    ModelIdentityMixin, Boto3Model
):
    """
    The base class for all boto3 models that get returned as the primary object
    from a boto3 operation.
    """

    #: The manager for this model
    manager_class: ClassVar[Type[Boto3ModelManager]]

    #: The boto3 session to use for this model.  This is set by the manager,
    #: and is used in relationships.  We have to use ``Any`` here because we
    #: pydantic complains vociferously if we use ``boto3.session.Session``.
    #: We exclude it from the model dump because it's not something that should
    #: be serialized.
    session: Optional[Any] = Field(default=None, exclude=True, frozen=False)

    @classproperty
    def objects(cls) -> Boto3ModelManager:  # noqa: N805
        """
        Get the manager for this model.

        Returns:
            The manager for this model.

        """
        return cls.manager_class()

    def save(self, **kwargs):
        """
        Save the model.
        """
        if self.pk:
            return self.objects.update(self, **kwargs)
        return self.objects.create(self, **kwargs)

    def delete(self):
        """
        Delete the model.
        """
        if not self.pk:
            msg = "Cannot delete a model that has not been saved."
            raise ValueError(msg)
        return self.manager.delete(self.pk)

    def set_session(self, session: boto3.session.Session) -> None:
        """
        Set the boto3 session for this model.

        Args:
            session: The boto3 session to use.

        Returns:
            The model instance.

        """
        self.session = session



from typing import TYPE_CHECKING, Dict, Final, Optional, Type, cast

import botocore.model

if TYPE_CHECKING:
    from botocraft.sync.service import ModelGenerator, ServiceGenerator


class AbstractShapeConverter:
    """
    The base class for all shape converters.
    """

    def __init__(self, shape_converter: "PythonTypeShapeConverter") -> None:
        self.service_generator = shape_converter.service_generator
        self.model_generator = shape_converter.model_generator
        self.shape_converter = shape_converter

    def model_exists(self, model_name: str) -> bool:
        """
        If we've already generated this model, return ``True``,
        else return ``False``.

        Args:
            model_name: the name of the model to check

        """
        model_def = self.model_generator.get_model_def(model_name)
        if model_def.alternate_name:
            model_name = model_def.alternate_name
        if model_name in self.model_generator.classes:
            # We generated this model earlier in this phase of service
            # generation
            return True
        if model_name in self.service_generator.classes:
            # We generated generated this model in a previous phase of service
            # generation
            return True
        if model_name in self.service_generator.service_def.models:
            # This model is belongs to this service, but we haven't generated it
            # yet.  Do so now.
            return False
        if model_name in self.service_generator.interface.models:  # noqa: SIM103
            # models we've generated so far in other services
            return True
        # This is some dependent model for this service. We'll generate it now.
        return False

    def import_line(self, model_name: str) -> Optional[str]:
        """
        Given ``model_name``, determine whether we've already
        generated this model somewhere else in the service, and
        """
        if model_name in self.service_generator.service_def.models:
            # This model is belongs to this service, but we haven't generated it
            # yet.  We don't need an import line, because the model lives
            # in this service.
            return None
        if model_name in self.service_generator.interface.models:
            import_path = self.service_generator.interface.models[model_name]
            return f"from {import_path} import {model_name}"
        return None

    def to_python(
        self, shape: botocore.model.Shape, quote: bool = False, name_only: bool = False
    ) -> str:
        """
        Converts the given shape ``shape`` to a python type,
        possibly using the model generator to generate
        pydantic models for nested structures.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator, if this
                is a structure shape.

        Returns:
            the python type for the shape

        """
        raise NotImplementedError


class StringShapeConverter(AbstractShapeConverter):
    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,  # noqa: ARG002
        name_only: bool = False,  # noqa: ARG002
    ) -> str:
        if shape.type_name == "string" or shape.name == "String":
            value = cast(botocore.model.StringShape, shape)
            if value.enum:
                contents = ", ".join([f"'{value}'" for value in value.enum])
                python_type = f"Literal[{contents}]"
            else:
                python_type = "str"
            return python_type
        msg = f"Not string type: {shape.type_name}"
        raise ValueError(msg)


class BooleanShapeConverter(AbstractShapeConverter):
    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,  # noqa: ARG002
        name_only: bool = False,  # noqa: ARG002
    ) -> str:
        if shape.type_name == "boolean":
            return "bool"
        msg = f"Not boolean type: {shape.type_name}"
        raise ValueError(msg)


class IntegerShapeConverter(AbstractShapeConverter):
    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,  # noqa: ARG002
        name_only: bool = False,  # noqa: ARG002
    ) -> str:
        if shape.type_name in ["integer", "long"]:
            return "int"
        msg = f"Not integer type: {shape.type_name}"
        raise ValueError(msg)


class DoubleShapeConverter(AbstractShapeConverter):
    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,  # noqa: ARG002
        name_only: bool = False,  # noqa: ARG002
    ) -> str:
        if shape.type_name == "double":
            return "float"
        msg = f"Not double type: {shape.type_name}"
        raise ValueError(msg)


class ListShapeConverter(AbstractShapeConverter):
    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,  # noqa: ARG002
        name_only: bool = False,
    ) -> str:
        """
        Convert a list shape to a python type.  If the list members are not
        strings, we will generate a pydantic model for the list members

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator, if this
                is a structure shape.

        Raises:
            ValueError: this is not a list shape

        Returns:
            a python type for the list

        """
        if shape.type_name != "list":
            msg = f"Not list type: {shape.type_name}"
            raise ValueError(msg)
        element_shape = cast(botocore.model.ListShape, shape).member
        inner_model_name = self.shape_converter.convert(
            element_shape, quote=True, name_only=name_only
        )
        return f"List[{inner_model_name}]"


class MapShapeConverter(AbstractShapeConverter):
    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,  # noqa: ARG002
        name_only: bool = False,
    ) -> str:
        """
        Convert a map shape to a python type.  We're assuming that
        the map key is always a string, and the value is always a
        string -- I haven't seen any other different examples.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator, if this
                is a structure shape.

        Raises:
            ValueError: this is not a map shape

        Returns:
            the python type for the map

        """
        if shape.type_name != "map":
            msg = f"Not map type: {shape.type_name}"
            raise ValueError(msg)
        shape = cast(botocore.model.MapShape, shape)
        value_type = self.shape_converter.convert(
            shape.value, quote=True, name_only=name_only
        )
        key_type = self.shape_converter.convert(
            shape.key, quote=True, name_only=name_only
        )
        return f"Dict[{key_type}, {value_type}]"


class StructureShapeConverter(AbstractShapeConverter):
    def to_python(
        self, shape: botocore.model.Shape, quote: bool = False, name_only: bool = False
    ) -> str:
        """
        Convert a structure shape to a python type.  This
        ends up always being a pydantic model.

        Side Effects:
            Might add a new pydantic class to the service python module.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator.

        Raises:
            ValueError: this is not a structure shape

        Returns:
            the python type for the structure

        """
        if shape.type_name != "structure":
            msg = f"Not structure type: {shape.type_name}"
            raise ValueError(msg)
        import_path: Optional[str] = None
        shape = cast(botocore.model.StructureShape, shape)
        model_def = self.model_generator.get_model_def(shape.name)
        model_name = shape.name
        if model_def.alternate_name:
            model_name = model_def.alternate_name
        if not name_only:
            if not self.model_exists(model_name):
                # We have not yet generated this model as a pydantic class
                model_name = self.model_generator.generate_model(
                    shape.name, model_shape=shape
                )
            else:  # noqa: PLR5501
                # We've already generated the model, so we just need to import it
                # if it's in another service, otherwise we can just use it

                # TODO: we have a problem here that if we want to import the model
                # from another service, but the model has an alternate name, we
                # won't find it this way because ``model_name`` here will just be
                # ``shape.name``
                if import_path := self.import_line(model_name):
                    self.model_generator.imports.add(import_path)
            if quote and not import_path:
                return f'"{model_name}"'
        elif quote:
            return f'"{model_name}"'
        return model_name


class TimestampShapeConverter(AbstractShapeConverter):
    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,  # noqa: ARG002
        name_only: bool = False,  # noqa: ARG002
    ) -> str:
        """
        Convert a timestamp shape to a python type.   This
        ends up always being a ``datetime.datetime`` object.

        Side Effects:
            This adds the ``datetime`` import to the model generator.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator.

        Raises:
            ValueError: this is not a timestamp shape

        Returns:
            the python type for the timestamp

        """
        if shape.type_name == "timestamp":
            return "datetime"
        self.model_generator.imports.add("from datetime import datetime")
        msg = f"Not timestamp type: {shape.type_name}"
        raise ValueError(msg)


class PythonTypeShapeConverter:
    CONVERTERS: Final[Dict[str, Type[AbstractShapeConverter]]] = {
        "string": StringShapeConverter,
        "boolean": BooleanShapeConverter,
        "integer": IntegerShapeConverter,
        "structure": StructureShapeConverter,
        "timestamp": TimestampShapeConverter,
        "double": DoubleShapeConverter,
        "list": ListShapeConverter,
        "map": MapShapeConverter,
    }

    def __init__(
        self, service_generator: "ServiceGenerator", model_generator: "ModelGenerator"
    ) -> None:
        #: The service generator.  We use this for accessing the
        #: model generator in case we need to build pydantic
        #: models out of a structure shape
        self.service_generator = service_generator
        self.model_generator = model_generator
        self.converters: Dict[str, AbstractShapeConverter] = {
            key: converter_class(self)
            for key, converter_class in self.CONVERTERS.items()
        }

    def convert(
        self, shape: botocore.model.Shape, quote: bool = False, name_only: bool = False
    ) -> str:
        """
        Convert a botocore shape to the appropriate python type.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the return value.
            name_only: if ``True``, we will return the name of the model and not
                create the model in the model generator.

        Raises:
            ValueError: Could not find a converter for the shape

        Returns:
            The python type for the shape to be used in type hints

        """
        for converter in self.converters.values():
            try:
                return converter.to_python(shape, quote=quote, name_only=name_only)
            except ValueError:  # noqa: PERF203
                pass
        msg = f"No converter for {shape.type_name}"
        raise ValueError(msg)
